/*
Virtual Machine
Copyright (C) 2005  Gianfranco Gallizia

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/***********************************************************************/
/*                                                                     */
/* Virtual Machine di Gianfranco Gallizia                              */
/*                                                                     */
/* Implementa una macchina virtuale che supporta il set di istruzioni  */
/*                                                                     */
/* 0	LDNA Load Numeric in Accumulator                               */
/* 1	LDAR Load Accumulator in Register                              */
/* 2	LDRA Load Register in Accumultor                               */
/* 3	ADDN Add Numeric to Accumulator                                */
/* 4	ADDR Add Register to Accumulator                               */
/* 5	SUBN Subtract Numeric to Accumulator                           */
/* 6	SUBR Subtract Register to Accumulator                          */
/* 7	MULN Multiply Numeric to Accumulator                           */
/* 8	MULR Multiply Register to Accumulator                          */
/* 9	DIVN Divide Accumulator by Numeric                             */
/* 10	DIVR Divide Accumulator by Register                            */
/* 11	ANDN Binary bit-to-bit AND between Accumulator and Numeric     */
/* 12	ANDR Binary bit-to-bit AND between Accumulator and Register    */
/* 13	ORNA Binary bit-to-bit OR between Accumulator and Numeric      */
/* 14	ORRA Binary bit-to-bit OR between Accumulator and Register     */
/* 15	JMPA Jump if Accumulator is 0                                  */
/* 16	JNZA Jump if Accumulator is NOT 0                              */
/* 17	JMPR Jump if Register is 0                                     */
/* 18	JNZR Jump if Register is NOT 0                                 */
/* 19	GOTO                                                           */
/* 20	HALT Break Execution                                           */
/* 21	GRAB Grab a 8-digit binary (in 0-255 format) number and put it */
/* 	in Accumulator                                                 */
/* 22	PRNT Print Accumulator                                         */
/*                                                                     */
/* Le istruzioni vanno inserite in formato numerico:                   */
/* GOTO120 -> 0x13000078 -> 318767224                                  */
/* Il byte piu' significativo contiene il numero dell'istruzione       */
/* Il byte meno significativo contiene l'operando dell'istruzione      */
/*                                                                     */
/***********************************************************************/

/*#define DEBUG /*Abilita la modalit�EBUG*/

#include <stdio.h>

int PC; /*Program Counter*/

int ACC; /*Accumulatore*/

int memory[256]={0}; /*Memoria*/

int IR; /*Instruction Register*/

int zero_f; /*flag di zero*/
int run_f=1; /*flag d'esecuzione*/

int loader(); /*loader per la modalit�interattiva*/

int floader(char *); /*file loader*/

int get_instr(int); /*recupera l'istruzione*/

int find_data(int); /*recupera l'operando*/

void interpreter(int); /*interprete*/

void execute(int,int); /*routine di esecuzione*/

/*Utilizzo: ./vm file oppure ./vm (modalit�interattiva)*/
/*In modalit�interattiva accetta solo input numerico*/

int main(int argc, char *argv[]) {

	int start;

	if (argc == 2) start = floader(argv[1]);
	else start = loader();
	if (start>=0)
		interpreter(start);
	else exit(1);

return 0;

}

void interpreter(int start_address) {

	int instr;
	int data;
	#ifdef DEBUG
	printf("Interprete avviato\n");
	#endif
	PC=start_address;
	
	while(run_f) {
	
		IR=memory[PC];
		PC++;
		instr = get_instr(IR);
		data = find_data(IR);
	
		execute(instr,data);
	
		if (ACC==0) zero_f=1; else zero_f=0; /*aggiorno il flag di zero*/
	
	}
}

int loader() {

#ifdef DEBUG
	printf("loader avviato\n");
#endif

	int input;
	int i=0;
	
	while(i<200) { /*Carica fino a 200 istruzioni*/
	
		scanf("%d",&input);
#ifdef DEBUG
	printf("input: %d\n",input);
#endif
		if(input == 666) break;
		memory[i]=input;
		i++;
	
	}
	
	return 0;

}

int get_instr(int instr) {

	int mask = 0xFF000000;
	int instruction = instr & mask;
	instruction /= (16777216);
#ifdef DEBUG
	printf("instruction: %d\n",instruction);
#endif
	return instruction;

}

int find_data(int instr) {

	int mask = 0x000000FF;
	int data = instr & mask;
#ifdef DEBUG
	printf("data: %d\n",data);
#endif
	return data;

}

void execute(int instr, int data) {

	switch(instr) {
	
		case 0 :	ACC=data;		/*LDNA*/
				break;
		case 1 :	memory[data]=ACC;	/*LDRA*/
				break;
		case 2 :	ACC=memory[data];	/*LDAR*/
				break;
		case 3 :	ACC+=data;		/*ADDN*/
				break;
		case 4 :	ACC+=memory[data];	/*ADDR*/
				break;
		case 5 :	ACC-=data;		/*SUBN*/
				break;
		case 6 :	ACC-=memory[data];	/*SUBR*/
				break;
		case 7 :	ACC*=data;		/*MULN*/
				break;
		case 8 :	ACC*=memory[data];	/*MULR*/
				break;
		case 9 :	ACC/=data;		/*DIVN*/
				break;
		case 10 :	ACC/=memory[data];	/*DIVN*/
				break;
		case 11 :	ACC&=data;		/*ANDN*/
				break;
		case 12 :	ACC&=memory[data];	/*ANDR*/
				break;
		case 13 :	ACC|=data;		/*ORNA*/
				break;
		case 14 :	ACC|=memory[data];	/*ORRA*/
				break;
		case 15 :	if(zero_f) PC++;	/*JMPA*/
				break;
		case 16 :	if(!zero_f) PC++;	/*JNZA*/
				break;
		case 17 :	if(!memory[data]) PC++;	/*JMPR*/
				break;
		case 18 :	if(memory[data]) PC++;	/*JNZR*/
				break;
		case 19 :	PC=data;		/*GOTO*/
				break;
		case 20 :	run_f=0;		/*HALT*/
				break;
		case 21 :	printf("> "); scanf("%d",&ACC); printf("\n");	/*GRAB*/
				break;
		case 22 :	printf("> %d\n",ACC);	/*PRNT*/
				break;
		default :	run_f=0;
	
	}
	

}

int floader(char *filename) {

#ifdef DEBUG
	printf("fileloader avviato\n");
#endif

	int input;
	int i=0;
	FILE *inputfile;
	
	if((inputfile = fopen(filename,"r"))==NULL) exit(1);
	
#ifdef DEBUG
	printf("file caricato\n");
#endif
	
	while(i<200) {
	
		fscanf(inputfile,"%d",&input);
#ifdef DEBUG
	printf("input: %d\n",input);
#endif
		if(input == 666) break;
		memory[i]=input;
		i++;
	
	}
	
	fclose(inputfile);
	
	return 0;

}
